// 2017.12.28 S.Koyama
#include <Wire.h>
#include <Adafruit_NeoPixel.h>
#include "MIDIUSB.h"
#include "pitchToFrequency.h"
#define Ttx		10000			// 少なくとも10秒毎に送信
#define CTL 6						// CDM7160 仕様変更
#define TSERVO	300			// SERVOは300msでdetatch
// ---------- (pin) --------
//							0				// MidiOUT
//							1				// MidiIN
#define SDA			2				// I2C SDA
#define SCL			3 			// I2C SCL
#define NEOpin	4				// NeoPixel pin
//							5#
#define BZZPIN	6				// BZZ(#A7)
//							7
//							8(A8)
//							9#(A9)	// servo
//							10#			// servo
//							14(MISO)// nRF24L
//							15(SCK) // nRF24L
//							16(MOSI)// nRF24L
#define A0			18
#define A1			19
#define A2			20			// nRF24L CSN
#define A3			21			// nRF24L CE
#define A4			22			// SHT31
#define A5			23			// SHT31
#define A6			24			// CDM7160
#define A7			6
#define TEMP		22			// SHT31
#define HUMI		23			// SHT31
#define CO2			24			// CDM7160
#define NPIN		25			// pin: 0-24
// ---------- (I2C) --------
#define SHT 		0x45		// SHT31
#define CDM			0x69		// CDM7160
#define CDM_CTL 0x01
#define CDM_ST1 0x02
#define CDM_DAL 0x03
#define CDM_DAH 0x04
#define BUSBUSY (digitalRead(SCL)==LOW||digitalRead(SDA)==LOW)
#define M1 0x64
#define M2 0x60
// ---------------------------
#define NPIXEL	12			// NeoPixel LED数
//---
//			INPUT					0
//			OUTPUT				1
//			INPUT_PULLUP	2
#define ANALOG_INPUT 	3
//---
Adafruit_NeoPixel pixels=Adafruit_NeoPixel(NPIXEL, NEOpin, NEO_GRB+NEO_KHZ800);
bool isSHT=false, isCDM=false;	// SHTもCDMも接続なし
byte pinStatus[NPIN], pixelFlag=0, d0,d1,d2,d3,d4,d5, lastbusy=0;
uint16_t Pins[NPIN], tbuf, hbuf, co2buf;
unsigned long currMillis, lastMillis, measTempMillis=0, measCO2Millis=0, servoMillis=0;
unsigned long lastTxTemp=0, lastTxHumi=0, lastTxCO2=0;
void servoWrite(uint8_t pin, uint8_t deg){	// deg: 0-180
//	int us=map(deg, 0,180, 544,2400);		// 544-2400 [ms]
	int us=map(deg, 0,180, 1000,2000);		// 1000-2000 [ms]
	if(pin==9 || pin==10){
		if(deg==0xFF){											// 停止
			TCCR1B&=0xE0;											// WGM1:00 CS1:000
			if(pin==9)	TCCR1A&=0x3C;					// COM1A:00 WGM1:00
			else				TCCR1A&=0xCC;					// COM1B:00 WGM1:00
		}else{															// 開始
			TCCR1B=TCCR1B&0xE0 | 0x18 | 0x02;	// WGM1:11 CS1:010(Ck/8)
			ICR1=40000;												// 20ms周期(0.5us×40000)
			if(pin==9){	OCR1A=us<<1;	TCCR1A=TCCR1A&0x3C|0x82;	}	// COM1A:10 WGM1:10
			else{				OCR1B=us<<1;	TCCR1A=TCCR1A&0xCC|0x22;	}	// COM1B:10 WGM1:10
		}
	}else	for(int t=0; t<300; t+=20){
		digitalWrite(pin, HIGH);	delayMicroseconds(us);
		digitalWrite(pin, LOW);		delayMicroseconds(20000-us);
	}
}
void sendMidi(byte cmd, byte val1, byte val2){
	midiEventPacket_t midicode={cmd>>4, cmd, val1, val2};
	MidiUSB.sendMIDI(midicode);
}
void setPinMode(byte pin, byte mode){
	if(pin>=NPIN || mode>ANALOG_INPUT) return;
	pinStatus[pin]=mode;	Pins[pin]=0;
	if(mode==OUTPUT){	digitalWrite(pin, LOW);	pinMode(pin, OUTPUT);}
	else if(mode==INPUT)											pinMode(pin, INPUT);
	else if(mode==INPUT_PULLUP)								pinMode(pin, INPUT_PULLUP);
	else if(pin>=A0 && pin<A4 && mode==ANALOG_INPUT)	pinMode(pin, INPUT);
}
void reset(){
	for(byte pin=0; pin<NPIN; pin++)	setPinMode(pin, OUTPUT);
	pixelFlag=0;
}
void noteOn(byte ch, byte pitch, byte velocity){
	tone(BZZPIN, pitchFrequency[pitch]);
}
void noteOff(byte channel, byte pitch, byte velocity){
	noTone(BZZPIN);
}
bool SHTbegin(){
	Wire.beginTransmission(SHT); Wire.write(0x30); Wire.write(0xA2);	// soft reset
	if(Wire.endTransmission()!=0) return false;
	return true;
}
bool CDMbegin(){
	Wire.beginTransmission(CDM); Wire.write(CDM_CTL); Wire.write(CTL);
	if(Wire.endTransmission()!=0) return false;
	delay(100);
	Wire.beginTransmission(CDM); Wire.write(CDM_ST1); Wire.endTransmission();
	while(BUSBUSY)	delay(10);				// バス開放を待つ
	Wire.requestFrom(CDM,1);					// 1バイト受信
	if(Wire.available()==0)	return false;
	return true;
}
void setup(){
//	Serial.begin(9600); while(!Serial);
	delay(100);
	Wire.begin();
	isSHT=SHTbegin();
	isCDM=CDMbegin();
	reset();
}
void analogSend(byte pin){					// A0-A3, A7
	if(pinStatus[pin]==ANALOG_INPUT){ // アナログ入力なら
		int val=analogRead(pin);				// 値を読み
		if(val!=Pins[pin]){							// 変化があれば送信
			byte apin=pin-A0;
			if(pin==6)	apin=7;						// A7: 6ピン
			sendMidi(0xE0, 0x40|((val>>4)&0x38)|apin, val&0x7F);
			MidiUSB.flush();							// 1110nnnn 01vvvaaa 0vvvvvvv
			Pins[pin]=val;								// 値を記憶
		}
	}
}
bool measTempHumi(){								// ***** 温湿度の測定 *****
	uint16_t t, h;
	Wire.beginTransmission(SHT);			// 値を読み
	Wire.write(0x24);Wire.write(0x00);
	if(Wire.endTransmission()!=0) return false;
	delay(15);
	Wire.requestFrom(SHT, 6);					// 6バイト受信
	if(Wire.available()!=6)	tbuf=hbuf=0;	// エラーなら値は0
	else{
		t=Wire.read(); t=(t<<8)|Wire.read(); Wire.read();// CRCは空読み
		h=Wire.read(); h=(h<<8)|Wire.read(); Wire.read();// CRCは空読み
		tbuf=(t>>4)&0x0FFF;							// 12ビットデータ
		hbuf=(h>>4)&0x0FFF;							// 12ビットデータ
	}
	return true;
}
uint16_t adjCO2(uint16_t co2){			// CO2値を
	if(co2>8191)	co2=8191;						// 最大8191とし(13ビット)
	return co2>>1;										// 1/2にして12ビットで表す(0～4095)
}
void measCO2(){											// ***** CO2の測定 *****
	uint16_t val;
	if(isCDM){
		Wire.beginTransmission(CDM); Wire.write(CDM_DAL); Wire.endTransmission();
		Wire.requestFrom(CDM,2);				// 2バイト受信
		if(Wire.available()!=2)	co2buf=0;	// エラーなら値は0
		else{
			val=Wire.read(); val|=((Wire.read()&0x7F)<<8);
			co2buf=val;									// 2秒毎のCO2を保存
		}
	}
	uint16_t co2=adjCO2(co2buf);		// 12ビットデータ
	if(co2!=Pins[A6]||(currMillis>=lastTxCO2+Ttx)){// tに変化があるか前回送信
		sendMidi(0xA0, (co2>>5)&0x7C|2, co2&0x7F);	// からTtx経過したら送信
		MidiUSB.flush();								// 1110nnnn 0vvvvv10 0vvvvvvv
		Pins[A6]=co2;										// 12ビット値を記憶
	}
}
void loop(){
// PCからのmidiメッセージを受信 ********************************************
	midiEventPacket_t rx=MidiUSB.read();
	byte cmd=rx.byte1&0xf0;
	if(rx.header!=0){									// PCからのmidiメッセージあり
		d0=rx.byte1; d1=rx.byte2; d2=rx.byte3;
		if(cmd==0x80){									// Midi NoteOff
			noteOff(d0&0x0F, d1, d2);			//channel, control, value
		}else if(cmd==0x90){						// Midi NoteOn
			noteOn (d0&0x0F, d1, d2); 		//channel, control, value
		}else if(cmd==0xA0){						// neoPixel
			if(pixelFlag==0){							// 1st byte
				d3=d0; d4=d1; d5=d2;	pixelFlag=1;
			}else{												// 2nd byte
				byte iPixel=d4&0x0F;
				byte R=((d4&0x40)<<1)|d5, G=((d4&0x20)<<2)|d1, B=((d4&0x10)<<3)|d2;
				pixels.setPixelColor(iPixel, R,G,B);	pixels.show();
				pixelFlag=0;
			}
		}else if(cmd==0xB0 && (d1&0x7C)==0x70){	// setPinMode, digitalWrite
			byte pin=((d1&0x03)<<3)|((d2&0x70)>>4);
			if((d2&0x0E)==0x0E){					// 1011nnnn 011100pp 0ppp111v
				if(pinStatus[pin]!=OUTPUT) setPinMode(pin, OUTPUT);
				digitalWrite(pin, d2&0x01);
			}else setPinMode(pin,d2&0x0F);// 1011nnnn 011100pp 0pppmmmm
		}else if(cmd==0xB0 && d1==0x79 && d2==0x00)	reset();	// RESET
		else if(cmd==0xE0){							// motor / servo / analogWrite
			uint8_t pin=d1&0x1F, dir;
			uint8_t val=((d1&0x20)<<2)|d2;// 0～255
			char speed=val;								// -128～127
			if(pinStatus[pin]!=OUTPUT) setPinMode(pin, OUTPUT);
			if((d1&0x40)==0x00){					// 1110nnnn 00vppppp 0vvvvvvv
				if(pin==30 || pin==31){			// *** Motor(M1またはM2) ***
					if(speed==0){			dir=3; val=0;	}												// STOP
					else if(speed>0){ dir=1; val=map( speed, 1,127, 6,63);}	// FWD
					else{							dir=2; val=map(-speed, 1,128, 6,63);}	// BWD
					Wire.beginTransmission(pin==30?M1:M2);
					Wire.write(0x00); Wire.write((val<<2)|dir);
					Wire.endTransmission();
				}else	servoWrite(pin, val);	// Servo
			}else analogWrite(pin, val);	// 1110nnnn 01vppppp 0vvvvvv
		}
	}	
// PCにmidiメッセージを送信 ********************************************
	currMillis=millis();
	if(currMillis-lastMillis>20){			// ***** 20ms毎に *****
		lastMillis+=20;
		for(byte pin=0; pin<NPIN; pin++)// すべてのピンについて
			if(pinStatus[pin]==INPUT || pinStatus[pin]==INPUT_PULLUP){	// 入力なら
				byte val=digitalRead(pin);	// 値を読み
				if(val!=Pins[pin]){					// 変化があれば送信
					sendMidi(0xB0, 0x70|(pin>>3), ((pin<<4)&0x70)|0x0E|val);
					MidiUSB.flush();					// 1011nnnn 011100pp 0ppp111v
					Pins[pin]=val;						// 値を記憶
				}
			}
		analogSend(A0);analogSend(A1);analogSend(A2);analogSend(A3);analogSend(A7);
	}else if((pinStatus[A4]==ANALOG_INPUT || pinStatus[A5]==ANALOG_INPUT) && isSHT
		&& currMillis>=measTempMillis+1000){	// SHTが有効なら1秒毎に
		measTempMillis=currMillis;
		measTempHumi();								// ***** 温湿度の測定 *****
		if(pinStatus[A4]==ANALOG_INPUT // tに変化があるか前回送信からTtx経過したら送信
			&& (tbuf!=Pins[A4]||(currMillis>=lastTxTemp+Ttx))){
			sendMidi(0xA0, (tbuf>>5)&0x7C, tbuf&0x7F);
			MidiUSB.flush();						// 1010nnnn 0vvvvv00 0vvvvvvv
			Pins[A4]=tbuf;							// 12ビット値を記憶
			lastTxTemp=currMillis;			// 送信時刻を記憶
		}
		if(pinStatus[A5]==ANALOG_INPUT	// hに変化があるか前回送信からTtx経過したら送信
			&& (hbuf!=Pins[A5]||(currMillis>=lastTxHumi+Ttx))){
			sendMidi(0xA0, (hbuf>>5)&0x7C|1, hbuf&0x7F);
			MidiUSB.flush();						// 1010nnnn 0vvvvv01 0vvvvvvv
			Pins[A5]=hbuf;							// 12ビット値を記憶
			lastTxHumi=currMillis;			// 送信時刻を記憶
		}
	}else if(pinStatus[A6]==ANALOG_INPUT && isCDM	// CDMが有効なら1秒毎に
		&& currMillis>=measCO2Millis+1000){
		byte cdmbusy;
		Wire.beginTransmission(CDM); Wire.write(CDM_ST1);	Wire.endTransmission();
		while(BUSBUSY)	delay(10);		// バス開放を待つ
		Wire.requestFrom(CDM,1);			// 1バイト受信
		cdmbusy=Wire.read()&0x80;
		if(lastbusy!=0 && cdmbusy==0)	measCO2();	// CDMのbusy完了でCO2の測定・送信
		lastbusy=cdmbusy;
	}
}
